/*
 * drivers/video/ep93xx/ep93xxfb.c
 *
 * Copyright 2009 Atmark Techno, Inc. All Rights Reserved.
 *
 * Based on the Cirrus Logic ep93xxfb driver.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

#include <linux/module.h>
/* #define DEBUG 1 */
#include <linux/kernel.h>
#include <linux/reboot.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/ioport.h>
#include <linux/interrupt.h>
#include <linux/dma-mapping.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/arch/fb.h>
#include <asm/arch/cx25871.h>
#include <asm/hardware.h>
#include <linux/platform_device.h>

#define FBDEV_NAME "ep93xxfb"

#define ep93xxfb_raster_outl(value, reg)		\
{							\
	outl(0xaa, EP93XX_RASTER_RASTER_SWLOCK);	\
	outl(value, reg);				\
}

#define ep93xxfb_syscon_outl(value, reg)	\
{						\
    outl(0xaa, EP93XX_SYSCON_SWLOCK);		\
    outl(value, reg);				\
}

unsigned int pseudo_palette[256];
unsigned long *cursor_data;
struct ep93xxfb_info epinfo;

int vout = -1;
char *vmode;
int vmodeidx = -1;
int depth = -1;
int pixel_mask;
int pixel_shift;
int pixel_format;

static int ep93xxfb_setcol(struct fb_info *info);

static int
write_reg(unsigned char ucRegAddr, unsigned char ucRegValue)
{
	unsigned long uiVal, uiDDR;
	unsigned char ucData, ucIdx, ucBit;
	unsigned long ulTimeout;

	uiVal = inl(EP93XX_GPIO_G_DATA);
	uiDDR = inl(EP93XX_GPIO_G_DIRECTION);
	uiVal |= (EP93XX_GPIO_G_EEPROM_DRIVE_DATOD |
		  EP93XX_GPIO_G_EEPROM_DRIVE_CLKOD);
	outl(uiVal, EP93XX_GPIO_G_DATA);

	udelay(EE_DELAY_USEC);

	uiDDR |= (EP93XX_GPIO_G_EEPROM_DRIVE_DATOD |
		  EP93XX_GPIO_G_EEPROM_DRIVE_CLKOD);
	outl(uiDDR, EP93XX_GPIO_G_DIRECTION);

	udelay(EE_DELAY_USEC);

	uiVal &= ~EP93XX_GPIO_G_EEPROM_DRIVE_DATOD;
	outl(uiVal, EP93XX_GPIO_G_DATA);

	udelay(EE_DELAY_USEC);

	uiVal &= ~EP93XX_GPIO_G_EEPROM_DRIVE_CLKOD;
	outl(uiVal, EP93XX_GPIO_G_DATA);

	udelay(EE_DELAY_USEC);

	for (ucIdx = 0; ucIdx < 3; ucIdx++) {
		if (ucIdx == 0)
			ucData = (unsigned char)CX25871_DEV_ADDRESS;
		else if (ucIdx == 1)
			ucData = ucRegAddr;
		else
			ucData = ucRegValue;

		for (ucBit = 0; ucBit < 8; ucBit++) {
			if (ucData & 0x80)
				uiVal |= EP93XX_GPIO_G_EEPROM_DRIVE_DATOD;
			else
				uiVal &= ~EP93XX_GPIO_G_EEPROM_DRIVE_DATOD;

			outl(uiVal, EP93XX_GPIO_G_DATA);

			udelay(EE_DELAY_USEC);

			outl(uiVal | EP93XX_GPIO_G_EEPROM_DRIVE_CLKOD,
			     EP93XX_GPIO_G_DATA);

			udelay(EE_DELAY_USEC);

			outl(uiVal, EP93XX_GPIO_G_DATA);

			udelay(EE_DELAY_USEC);

			ucData <<= 1;
		}

		uiDDR &= ~EP93XX_GPIO_G_EEPROM_DRIVE_DATOD;
		outl(uiDDR, EP93XX_GPIO_G_DIRECTION);

		udelay(EE_DELAY_USEC);

		outl(uiVal | GPIOG_EECLK, EP93XX_GPIO_G_DATA);
		udelay(EE_DELAY_USEC);

		ulTimeout = 0;
		while (inl(EP93XX_GPIO_G_DATA) & EP93XX_GPIO_G_EEPROM_DRIVE_DATOD) {
			udelay( EE_DELAY_USEC );
			ulTimeout++;
			if (ulTimeout > EE_READ_TIMEOUT )
				return 1;
		}

		outl(uiVal, EP93XX_GPIO_G_DATA);

		udelay( EE_DELAY_USEC );

		uiDDR |= EP93XX_GPIO_G_EEPROM_DRIVE_DATOD;
		outl(uiDDR, EP93XX_GPIO_G_DIRECTION);

		udelay(EE_DELAY_USEC);
	}

	uiVal &= ~EP93XX_GPIO_G_EEPROM_DRIVE_DATOD;
	outl(uiVal, EP93XX_GPIO_G_DATA);

	udelay(EE_DELAY_USEC);

	uiVal |= EP93XX_GPIO_G_EEPROM_DRIVE_CLKOD;
	outl(uiVal, EP93XX_GPIO_G_DATA);

	udelay(EE_DELAY_USEC);

	uiVal |= EP93XX_GPIO_G_EEPROM_DRIVE_DATOD;
	outl(uiVal, EP93XX_GPIO_G_DATA);

	udelay(EE_DELAY_USEC);

	return 0;
}

static int
cx25871_reboot(struct notifier_block *this, unsigned long event, void *x)
{
	switch (event)
	{
	case SYS_HALT:
	case SYS_POWER_OFF:
	case SYS_RESTART:
	        write_reg(0xba, 0x80);
	default:
		return NOTIFY_DONE;
	}
}

static struct notifier_block cx25871_notifier = {
	cx25871_reboot,
	NULL,
	5
};

static void
cx25871_on( unsigned char value )
{
	write_reg(0x30,0x00);
	write_reg(0xBA, 0x20 );
}

static void
cx25871_off( unsigned char value )
{
	write_reg(0xBA, 0x30);
	write_reg(0xC4, 0x00);
	write_reg(0x30, 0x02);
}

static void
cx25871_config( unsigned char value )
{
	write_reg(0xBA, CX25871_REGxBA_SRESET);
	mdelay(1);
	write_reg(0xB8, value);
	write_reg(0xBA, CX25871_REGxBA_SLAVER | CX25871_REGxBA_DACOFF);
	write_reg(0xC6,(CX25871_REGxC6_INMODE_MASK & 0x3));
	write_reg(0xC4, CX25871_REGxC4_EN_OUT);
	write_reg(0xBA, CX25871_REGxBA_SLAVER );
}

static void
philips_lb064v02_on(unsigned char value)
{
	pr_debug("%s\n", __FUNCTION__);

	outl(inl(EP93XX_GPIO_A_DIRECTION) | 2, EP93XX_GPIO_A_DIRECTION);
	outl(inl(EP93XX_GPIO_A_DATA) | 2, EP93XX_GPIO_A_DATA);
}

static void
philips_lb064v02_off(unsigned char value)
{
	pr_debug("%s\n", __FUNCTION__);

	outl(inl(EP93XX_GPIO_A_DATA) & ~2, EP93XX_GPIO_A_DATA);
}

static unsigned long
ep93xx_get_pll_frequency(unsigned long pll)
{
	unsigned long fb1, fb2, ipd, ps, freq;

	if (pll == 1)
		pll = inl(EP93XX_SYSCON_CLOCK_SET1);
	else if (pll == 2)
		pll = inl(EP93XX_SYSCON_CLOCK_SET2);
	else
		return 0;

	ps = ((pll & EP93XX_SYSCON_CLOCK_SET1_PLL1_PS_MASK) >>
		EP93XX_SYSCON_CLOCK_SET1_PLL1_PS_SHIFT);
	fb1 = ((pll & EP93XX_SYSCON_CLOCK_SET1_PLL1_X1FBD1_MASK) >>
	       EP93XX_SYSCON_CLOCK_SET1_PLL1_X1FBD1_SHIFT);
	fb2 = ((pll & EP93XX_SYSCON_CLOCK_SET1_PLL1_X2FBD2_MASK) >>
	       EP93XX_SYSCON_CLOCK_SET1_PLL1_X2FBD2_SHIFT);
	ipd = ((pll & EP93XX_SYSCON_CLOCK_SET1_PLL1_X2IPD_MASK) >>
	       EP93XX_SYSCON_CLOCK_SET1_PLL1_X2IPD_SHIFT);

	freq = (((0x00e10000 * (fb1+1)) / (ipd+1)) * (fb2+1)) >> ps;

	return freq;
}

static int
ep93xx_get_max_video_clk(void)
{
	unsigned long f,freq = 0;

	freq = 14745600 / 4;
	f = ep93xx_get_pll_frequency(1) / 4;
	if ( f > freq )
		freq = f;
	f = ep93xx_get_pll_frequency(2) / 4;
	if ( f > freq )
		freq = f;

	return freq;
}

static int
ep93xx_set_video_div(unsigned long freq)
{
	unsigned long pdiv = 0;
	unsigned long div  = 0;
	unsigned long psel = 0;
	unsigned long esel = 0;
	unsigned long err, f, i, j, k;

	err = -1;

	for (i = 0; i < 3; i++) {
		if (i == 0)
			f = 14745600 * 2;
		else if (i == 1)
			f = ep93xx_get_pll_frequency(1) * 2;
		else
			f = ep93xx_get_pll_frequency(2) * 2;

		for (j = 4; j <= 6; j++) {
			k = f / (freq * j);
			if (k < 2)
				continue;

			if (abs(((f / (j * k))) - freq ) < err ) {
				pdiv = j - 3;
				div = k;
				psel = (i == 2) ? 1 : 0;
				esel = (i == 0) ? 0 : 1;
				err = (f / (j * k)) - freq;
			}
		}
	}

	if (err == -1)
		return -1;

	ep93xxfb_syscon_outl(EP93XX_SYSCON_VIDEO_CLOCK_VENA |
			     (esel ? EP93XX_SYSCON_VIDEO_CLOCK_ESEL : 0) |
			     (psel ? EP93XX_SYSCON_VIDEO_CLOCK_PSEL : 0) |
			     (pdiv << EP93XX_SYSCON_VIDEO_CLOCK_PDIV_SHIFT) |
			     (div << EP93XX_SYSCON_VIDEO_CLOCK_VDIV_SHIFT),
			     EP93XX_SYSCON_VIDEO_CLOCK);

	return freq + err;
}

static int
ep93xxfb_cursor(struct fb_info *info, struct ep93xx_cursor *cursor)
{
	unsigned long x, y, save;

	if (cursor->flags & CURSOR_OFF)
		outl(inl(EP93XX_RASTER_CURSORXYLOC) &
		     ~EP93XX_RASTER_CURSORXYLOC_CEN,
		     EP93XX_RASTER_CURSORXYLOC);

	if (cursor->flags & CURSOR_SETSHAPE) {
		if (copy_from_user(cursor_data, cursor->data,
				   cursor->width * cursor->height / 4))
			return -EFAULT;

		save = inl(EP93XX_RASTER_CURSORXYLOC);
		outl(save & ~EP93XX_RASTER_CURSORXYLOC_CEN,
		     EP93XX_RASTER_CURSORXYLOC);

		outl(virt_to_phys(cursor_data), EP93XX_RASTER_CURSOR_ADR_START);
		outl(virt_to_phys(cursor_data), EP93XX_RASTER_CURSOR_ADR_RESET);
		outl(((cursor->width - 1) & 0x30) << 4 |
		     ((cursor->height - 1) << 2) |
		     ((cursor->width - 1) >> 4), EP93XX_RASTER_CURSORSIZE);
		outl(save, EP93XX_RASTER_CURSORXYLOC);

	}

	if (cursor->flags & CURSOR_SETCOLOR) {
		outl(cursor->color1, EP93XX_RASTER_CURSORCOLOR1);
		outl(cursor->color2, EP93XX_RASTER_CURSORCOLOR2);
		outl(cursor->blinkcolor1, EP93XX_RASTER_CURSORBLINK1);
		outl(cursor->blinkcolor2, EP93XX_RASTER_CURSORBLINK2);
	}

	if (cursor->flags & CURSOR_BLINK) {
		if (cursor->blinkrate)
			outl(EP93XX_RASTER_CURSORBLINK_EN | cursor->blinkrate,
			     EP93XX_RASTER_CURSORBLINK);
		else
			outl(0x0000000FF, EP93XX_RASTER_CURSORBLINK);
	}

	if (cursor->flags & CURSOR_MOVE) {
		x = (inl(EP93XX_RASTER_HACTIVESTRTSTOP) & 0x000007ff) -
			cursor->dx - 2;
		y = (inl(EP93XX_RASTER_VACTIVESTRTSTOP) & 0x000007ff) - cursor->dy;
		outl((inl(EP93XX_RASTER_CURSORXYLOC) &
		      EP93XX_RASTER_CURSORXYLOC_CEN) |
		     (y << 16) | x, EP93XX_RASTER_CURSORXYLOC);
	}

	if (cursor->flags & CURSOR_ON)
		outl(inl(EP93XX_RASTER_CURSORXYLOC) |
		     EP93XX_RASTER_CURSORXYLOC_CEN,
		     EP93XX_RASTER_CURSORXYLOC);

	return 0;
}



static inline void
ep93xxfb_palette_write(u_int regno, u_int red, u_int green,
                       u_int blue, u_int trans)
{
	unsigned int cont, i, pal;

	pr_debug("%s  - enter\n", __FUNCTION__);

	pal = ((red & 0xFF00) << 8) | (green & 0xFF00) | ((blue & 0xFF00) >> 8);
	pseudo_palette[regno] = pal;
	outl( pal, ( EP93XX_RASTER_COLOR_LUT + ( regno << 2 )));
	cont = inl(EP93XX_RASTER_LUTCONT);

	if (( cont & EP93XX_RASTER_LUTCONT_SWTCH &&
	      cont & EP93XX_RASTER_LUTCONT_STAT ) ||
	    ( !( cont & EP93XX_RASTER_LUTCONT_SWTCH ) &&
	      !( cont & EP93XX_RASTER_LUTCONT_STAT))) {
		for (i = 0; i < 256; i++)
			outl( pseudo_palette[i],
			      ( EP93XX_RASTER_COLOR_LUT + ( i << 2 )) );

		outl(cont ^ EP93XX_RASTER_LUTCONT_SWTCH,
		      EP93XX_RASTER_LUTCONT);
	}
}

static int
ep93xxfb_setcolreg(unsigned regno, unsigned red, unsigned green,
		   unsigned blue, unsigned transp, struct fb_info *info)
{
#define CNVT_TOHW(val,width) ((((val)<<(width))+0x7FFF-(val))>>16)

	switch (info->fix.visual) {
	case FB_VISUAL_PSEUDOCOLOR:
		ep93xxfb_palette_write(regno, red, green, blue, transp);
		break;
	case FB_VISUAL_TRUECOLOR:
		if (regno >= 16)
			return 1;

		red = CNVT_TOHW(red, info->var.red.length);
		green = CNVT_TOHW(green, info->var.green.length);
		blue = CNVT_TOHW(blue, info->var.blue.length);
		transp = CNVT_TOHW(transp, info->var.transp.length);
		((u32 *)(info->pseudo_palette))[regno] =
			(red << info->var.red.offset) |
			(green << info->var.green.offset) |
			(blue << info->var.blue.offset) |
			(transp << info->var.transp.offset);
		break;
        case FB_VISUAL_DIRECTCOLOR:
		red = CNVT_TOHW(red, 8);
		green = CNVT_TOHW(green, 8);
		blue = CNVT_TOHW(blue, 8);
		transp = CNVT_TOHW(transp, 8);
		break;
	}
#undef CNVT_TOHW

	return 0;
}

static int
ep93xx_pan_display(struct fb_var_screeninfo *var, struct fb_info *info)
{
	pr_debug("%s  - enter\n", __FUNCTION__);

	if (var->yoffset < 0 ||
	    var->yoffset + var->yres > info->var.yres_virtual ||
	    var->xoffset)
		return -EINVAL;

	outl(epinfo.fb_phys + info->fix.line_length * var->yoffset,
	     EP93XX_RASTER_VIDSCRNPAGE);

	info->var.xoffset = var->xoffset;
	info->var.yoffset = var->yoffset;

	pr_debug("%s  - exit\n", __FUNCTION__);

	return 0;
}

static int
ep93xxfb_check_var(struct fb_var_screeninfo *var, struct fb_info *info)
{
	unsigned long pclk;

	pr_debug("%s  - enter\n", __FUNCTION__);

	if (vout != 0)
		return -EINVAL;

	if ((var->vmode & FB_VMODE_MASK) != FB_VMODE_NONINTERLACED) {
		pr_err("ep93xxfb_check_var - unsupported video mode\n");
		return -EINVAL;
	}

	if (((var->xres * var->yres * var->bits_per_pixel) / 8) > MAX_FBMEM_SIZE ) {
		pr_err("ep93xxfb_check_var - memory error\n");
		return -ENOMEM;
	}

	if (((var->xres_virtual * var->yres_virtual *
	      var->bits_per_pixel) / 8) > MAX_FBMEM_SIZE ) {
		pr_err("ep93xxfb_check_var - memory error\n");
		return -ENOMEM;
	}

	pclk = 1000 * (1000000000 / var->pixclock);

	if (pclk > ep93xx_get_max_video_clk()) {
		pr_err("ep93xxfb_check_var - pixel clock error %lu\n", pclk);
		return -EINVAL;
	}

	if (var->xres_virtual != var->xres)
		var->xres_virtual = var->xres;
	if (var->yres_virtual < var->yres)
		var->yres_virtual = var->yres;

	if (var->xoffset < 0)
		var->xoffset = 0;
	if (var->yoffset < 0)
		var->yoffset = 0;

	switch (var->bits_per_pixel) {
	case 8:
		break;
        case 16:
		break;
        case 24:
		break;
        case 32:
		break;
	default:
		return -EINVAL;
	}

	pr_debug("%s  - exit\n", __FUNCTION__);

	return 0;
}

static int
ep93xxfb_blank(int blank_mode,struct fb_info *info)
{
	unsigned long attribs;

	pr_debug("%s  - enter\n", __FUNCTION__);

	attribs = inl(EP93XX_RASTER_VIDEOATTRIBS);

	attribs &= ~EP93XX_RASTER_VIDEOATTRIBS_SDSEL_MASK;
	attribs |= CONFIG_EP93XX_VIDEOATTRIBS_SDSEL;

	if (blank_mode) {
		if (epinfo.off)
			(epinfo.off)(0);

		ep93xxfb_raster_outl(attribs & ~(EP93XX_RASTER_VIDEOATTRIBS_DATAEN |
					  EP93XX_RASTER_VIDEOATTRIBS_SYNCEN |
					  EP93XX_RASTER_VIDEOATTRIBS_PCLKEN |
					  EP93XX_RASTER_VIDEOATTRIBS_EN),
			      EP93XX_RASTER_VIDEOATTRIBS);
	} else {
		if (epinfo.clk_src == CLK_INTERNAL)
			attribs |= EP93XX_RASTER_VIDEOATTRIBS_PCLKEN;
		else
			attribs &= ~EP93XX_RASTER_VIDEOATTRIBS_PCLKEN;

		ep93xxfb_raster_outl(attribs | EP93XX_RASTER_VIDEOATTRIBS_DATAEN |
			      EP93XX_RASTER_VIDEOATTRIBS_SYNCEN |
			      EP93XX_RASTER_VIDEOATTRIBS_EN,
			      EP93XX_RASTER_VIDEOATTRIBS);

		if (epinfo.configure)
			(epinfo.configure)(epinfo.automods);

		if (epinfo.on)
			(epinfo.on)(0);
	}

	return 0;
}

static int
ep93xxfb_setclk(void)
{
	unsigned long calc_clk,act_clk;

	if (epinfo.clk_src == CLK_INTERNAL) {
		ep93xxfb_syscon_outl(inl(EP93XX_SYSCON_DEVICE_CONFIG) &
				     ~EP93XX_SYSCON_DEVICE_CONFIG_EXVC,
				     EP93XX_SYSCON_DEVICE_CONFIG);

		calc_clk = epinfo.refresh;
		act_clk = ep93xx_set_video_div(calc_clk);
		if (act_clk == -1)
			return -ENODEV;

		epinfo.refresh = act_clk;
		epinfo.pixclock = 1000000000 / (act_clk / 1000);
	} else {
		ep93xxfb_syscon_outl(0, EP93XX_SYSCON_VIDEO_CLOCK);

		ep93xxfb_syscon_outl(inl(EP93XX_SYSCON_DEVICE_CONFIG) |
				     EP93XX_SYSCON_DEVICE_CONFIG_EXVC,
				     EP93XX_SYSCON_DEVICE_CONFIG);
	}

	return 0;
}

static void
ep93xxfb_timing_signal_generation(void)
{
	unsigned long vlinestotal;
	unsigned long vsyncstart, vsyncstop;
	unsigned long vactivestart, vactivestop;
	unsigned long vblankstart, vblankstop;
	unsigned long vclkstart, vclkstop;

	unsigned long hclkstotal;
	unsigned long hsyncstart, hsyncstop;
	unsigned long hactivestart, hactivestop;
	unsigned long hblankstart, hblankstop;
	unsigned long hclkstart, hclkstop;

	pr_debug("%s - enter\n", __FUNCTION__);

	vlinestotal = epinfo.ytotal - 1;
	vsyncstart = vlinestotal;
	vsyncstop = vlinestotal - epinfo.ysync;
	vblankstart = vlinestotal - epinfo.ysync - epinfo.ybp;
	vblankstop = epinfo.yfp - 1;
	vactivestart = vblankstart;
	vactivestop = vblankstop;
	vclkstart = vlinestotal;
	vclkstop = vlinestotal + 1;

	hclkstotal = epinfo.xtotal - 1;
	hsyncstart = hclkstotal;
	hsyncstop = hclkstotal - epinfo.xsync;
	hblankstart = hclkstotal - epinfo.xsync - epinfo.xbp;
	hblankstop = epinfo.xfp - 1;
	hactivestart = hblankstart;
	hactivestop = hblankstop;
	hclkstart = hclkstotal ;
	hclkstop = hclkstotal ;

	ep93xxfb_raster_outl(0, EP93XX_RASTER_VIDEOATTRIBS);

	ep93xxfb_raster_outl(vlinestotal, EP93XX_RASTER_VLINESTOTAL);
	ep93xxfb_raster_outl(vsyncstart + (vsyncstop << 16),
		      EP93XX_RASTER_VSYNCSTRTSTOP);
	ep93xxfb_raster_outl(vactivestart + (vactivestop << 16),
		      EP93XX_RASTER_VACTIVESTRTSTOP);
	ep93xxfb_raster_outl(vblankstart + (vblankstop << 16),
		      EP93XX_RASTER_VBLANKSTRTSTOP);
	ep93xxfb_raster_outl(vclkstart + (vclkstop << 16),
		      EP93XX_RASTER_VCLKSTRTSTOP);

	ep93xxfb_raster_outl(hclkstotal , EP93XX_RASTER_HCLKSTOTAL);
	ep93xxfb_raster_outl(hsyncstart + (hsyncstop << 16),
		      EP93XX_RASTER_HSYNCSTRTSTOP);
	ep93xxfb_raster_outl(hactivestart + (hactivestop << 16),
		      EP93XX_RASTER_HACTIVESTRTSTOP);
	ep93xxfb_raster_outl(hblankstart + (hblankstop << 16),
		      EP93XX_RASTER_HBLANKSTRTSTOP);
	ep93xxfb_raster_outl(hclkstart + (hclkstop << 16) ,
		      EP93XX_RASTER_HCLKSTRTSTOP);

	ep93xxfb_raster_outl(0, EP93XX_RASTER_LINECARRY);
}

static
void ep93xxfb_set_pixelmode(int bpp)
{
	unsigned long tmpdata = 0;

	pr_debug("%s %dbpp - enter\n", __FUNCTION__, bpp);

	switch(bpp) {
	case 8:
		tmpdata = EP93XX_RASTER_PIXELMODE_P_8BPP |
			EP93XX_RASTER_PIXELMODE_S_1PPCMAPPED |
			EP93XX_RASTER_PIXELMODE_C_LUT;
		epinfo.pixformat = PIXEL_FORMAT_8;
		break;
	case 16:
		tmpdata = EP93XX_RASTER_PIXELMODE_P_16BPP |
			EP93XX_RASTER_PIXELMODE_S_1PPCMAPPED |
			EP93XX_RASTER_PIXELMODE_C_565;
		epinfo.pixformat = PIXEL_FORMAT_16;
		break;
	case 24:
		tmpdata = EP93XX_RASTER_PIXELMODE_P_24BPP |
			EP93XX_RASTER_PIXELMODE_S_1PPC |
			EP93XX_RASTER_PIXELMODE_C_888;
		epinfo.pixformat = PIXEL_FORMAT_24;
		break;
        case 32:
		tmpdata = EP93XX_RASTER_PIXELMODE_P_32BPP |
			EP93XX_RASTER_PIXELMODE_S_1PPC |
			EP93XX_RASTER_PIXELMODE_C_888;
		epinfo.pixformat = PIXEL_FORMAT_32;
		break;
	default:
		break;
	}

	outl(tmpdata, EP93XX_RASTER_PIXELMODE);
}

static int
ep93xxfb_set_par(struct fb_info *info)
{
	unsigned long attribs;
	int ret;

	pr_debug("%s  - enter\n", __FUNCTION__);

	if (ep93xxfb_check_var(&info->var,info) < 0)
		return -EINVAL;

	ret = ep93xxfb_setcol(info);
	if (ret < 0)
		return ret;

	info->fix.line_length = info->var.xres *
		info->var.bits_per_pixel / 8;

	ep93xxfb_blank(1, info);

	epinfo.xres = info->var.xres;
	epinfo.xsync = info->var.hsync_len;
	epinfo.xfp = info->var.right_margin;
	epinfo.xbp = info->var.left_margin;
	epinfo.xtotal = epinfo.xres + epinfo.xsync + epinfo.xfp + epinfo.xbp;

	epinfo.yres = info->var.yres;
	epinfo.ysync = info->var.vsync_len;
	epinfo.yfp = info->var.lower_margin;
	epinfo.ybp = info->var.upper_margin;
	epinfo.ytotal = epinfo.yres + epinfo.ysync + epinfo.yfp + epinfo.ybp;

	epinfo.pixclock = info->var.pixclock;
	epinfo.refresh = 1000 * (1000000000 / info->var.pixclock);

	if (epinfo.refresh > ep93xx_get_max_video_clk())
		epinfo.refresh = ep93xx_get_max_video_clk();
	epinfo.bpp = info->var.bits_per_pixel;

	ep93xxfb_setclk();
	ep93xxfb_timing_signal_generation();

	// set video memory parameters
	outl(epinfo.fb_phys, EP93XX_RASTER_VIDSCRNPAGE);
	outl(epinfo.yres , EP93XX_RASTER_SCRNLINES);
	outl(((epinfo.xres * epinfo.bpp) / 32) - 1, EP93XX_RASTER_LINELENGTH);
	outl((epinfo.xres * epinfo.bpp) / 32, EP93XX_RASTER_VLINESTEP);

	// set pixel mode
	ep93xxfb_set_pixelmode(epinfo.bpp);

	attribs = 0;
	attribs |= CONFIG_EP93XX_VIDEOATTRIBS_SDSEL;

	attribs |= EP93XX_RASTER_VIDEOATTRIBS_INVCLK;
	if (info->var.sync & FB_SYNC_HOR_HIGH_ACT)
		attribs |= EP93XX_RASTER_VIDEOATTRIBS_HSPOL;
	if (info->var.sync & FB_SYNC_VERT_HIGH_ACT)
		attribs |= EP93XX_RASTER_VIDEOATTRIBS_VCPOL;

	ep93xxfb_raster_outl(attribs, EP93XX_RASTER_VIDEOATTRIBS);
	ep93xxfb_blank(0 , info);

	pr_debug("%s  - exit\n", __FUNCTION__);

	return 0;
}

static int
ep93xxfb_mmap(struct fb_info *info,struct vm_area_struct *vma)
{
	unsigned long off, start, len;

	pr_debug("%s  - enter\n", __FUNCTION__);

	off = vma->vm_pgoff << PAGE_SHIFT;
	start = info->fix.smem_start;
	len = PAGE_ALIGN(start & ~PAGE_MASK) + info->fix.smem_len;
	start &= PAGE_MASK;

	if ((vma->vm_end - vma->vm_start + off) > len)
		return -EINVAL;

	off += start;
	vma->vm_pgoff = off >> PAGE_SHIFT;

	vma->vm_flags |= VM_IO;
	vma->vm_page_prot = pgprot_writecombine(vma->vm_page_prot);

	if (io_remap_pfn_range(vma, vma->vm_start, off >> PAGE_SHIFT,
			       vma->vm_end - vma->vm_start, vma->vm_page_prot)) {
		pr_err("ep93xxfb_mmap error\n");
		return -EAGAIN;
	}

	pr_debug("%s  - exit\n", __FUNCTION__);

	return 0;
}

static int
ep93xxfb_setcol(struct fb_info *info)
{
	pr_debug("%s\n", __FUNCTION__);

	switch (info->var.bits_per_pixel) {
	case 8:
		info->var.red.length = 8;
		info->var.green.length = 8;
		info->var.blue.length = 8;
		info->fix.visual = FB_VISUAL_PSEUDOCOLOR;
		break;
	case 16:
		info->var.red.offset = 11;
		info->var.red.length = 5;
		info->var.green.offset = 5;
		info->var.green.length = 6;
		info->var.blue.offset = 0;
		info->var.blue.length = 5;
		info->fix.visual = FB_VISUAL_TRUECOLOR;
		break;
	case 24:
		info->var.red.length = 8;
		info->var.blue.length = 8;
		info->var.green.length = 8;
		info->var.red.offset = 16;
		info->var.green.offset = 8;
		info->var.blue.offset = 0;
		info->fix.visual = FB_VISUAL_TRUECOLOR;
		break;
	case 32:
		info->var.red.length = 8;
		info->var.blue.length = 8;
		info->var.green.length = 8;
		info->var.transp.length = 0;
		info->var.red.offset = 16;
		info->var.green.offset = 8;
		info->var.blue.offset = 0;
		info->var.transp.offset = 0;
		info->fix.visual = FB_VISUAL_TRUECOLOR;
		break;
	default:
		return -EINVAL;
	}

	return 0;
}

static void
ep93xxfb_get_par(struct fb_info *info)
{
	epinfo.configure = NULL;
	epinfo.on = NULL;
	epinfo.off = NULL;

	pr_debug("%s - enter\n", __FUNCTION__);

	switch (vout) {
	case LCD_OUT:
		epinfo.on = philips_lb064v02_on;
		epinfo.off = philips_lb064v02_off;

	case CRT_OUT:
		epinfo.xres = ep93xxfb_crtmodes[vmodeidx].hres;
		epinfo.xsync = ep93xxfb_crtmodes[vmodeidx].hsync;
		epinfo.xfp = ep93xxfb_crtmodes[vmodeidx].hfp;
		epinfo.xbp = ep93xxfb_crtmodes[vmodeidx].hbp;
		epinfo.xtotal = epinfo.xres + epinfo.xsync +
			epinfo.xfp + epinfo.xbp;

		epinfo.yres = ep93xxfb_crtmodes[vmodeidx].vres;
		epinfo.ysync = ep93xxfb_crtmodes[vmodeidx].vsync;
		epinfo.yfp = ep93xxfb_crtmodes[vmodeidx].vfp;
		epinfo.ybp = ep93xxfb_crtmodes[vmodeidx].vbp;
		epinfo.ytotal = epinfo.yres + epinfo.ysync +
        	        epinfo.yfp + epinfo.ybp;

		epinfo.refresh = epinfo.xtotal * epinfo.ytotal *
			ep93xxfb_crtmodes[vmodeidx].refresh;
		epinfo.pixclock = 1000000000 / ( epinfo.refresh / 1000);
		epinfo.bpp = depth;

		epinfo.clk_src = ep93xxfb_crtmodes[vmodeidx].clk_src;
		epinfo.clk_edge = ep93xxfb_crtmodes[vmodeidx].clk_edge;
		epinfo.pol_blank = ep93xxfb_crtmodes[vmodeidx].pol_blank;
		epinfo.pol_xsync = ep93xxfb_crtmodes[vmodeidx].pol_hsync;
		epinfo.pol_ysync = ep93xxfb_crtmodes[vmodeidx].pol_vsync;
		break;

	case NTSC_OUT:
		epinfo.xres = cx_ntscmodes[vmodeidx].hres;
		epinfo.xsync = 4;
		epinfo.xbp = cx_ntscmodes[vmodeidx].hblank;
		epinfo.xfp =  cx_ntscmodes[vmodeidx].hclktotal - epinfo.xbp -
			cx_ntscmodes[vmodeidx].hres - 4;
		epinfo.xtotal = cx_ntscmodes[vmodeidx].hclktotal;

		epinfo.yres = cx_ntscmodes[vmodeidx].vres;
		epinfo.ysync = 2;
		epinfo.ybp = cx_ntscmodes[vmodeidx].vblank;
		epinfo.yfp = cx_ntscmodes[vmodeidx].vclktotal - epinfo.ybp -
			cx_ntscmodes[vmodeidx].vres - 2;
		epinfo.ytotal = cx_ntscmodes[vmodeidx].vclktotal;
		epinfo.bpp = depth;

		epinfo.clk_src = CLK_EXTERNAL;
		epinfo.automods = cx_ntscmodes[vmodeidx].automode;
		epinfo.clk_edge = EDGE_FALLING;
		epinfo.pol_blank = POL_LOW;
		epinfo.pol_xsync = POL_LOW,
			epinfo.pol_ysync = POL_LOW;

		epinfo.refresh = cx_ntscmodes[vmodeidx].clkfrequency;
		epinfo.pixclock = 1000000000 / ( epinfo.refresh / 1000);

		epinfo.configure = cx25871_config;
		epinfo.on = cx25871_on;
		epinfo.off = cx25871_off;
		break;

	case PAL_OUT:
		epinfo.xres = cx_palmodes[vmodeidx].hres;
		epinfo.xsync = 4;
		epinfo.xbp = cx_palmodes[vmodeidx].hblank;
		epinfo.xfp =  cx_palmodes[vmodeidx].hclktotal - epinfo.xbp -
			cx_palmodes[vmodeidx].hres - 4;
		epinfo.xtotal = cx_palmodes[vmodeidx].hclktotal;

		epinfo.yres = cx_palmodes[vmodeidx].vres;
		epinfo.ysync = 2;
		epinfo.ybp = cx_palmodes[vmodeidx].vblank;
		epinfo.yfp = cx_palmodes[vmodeidx].vclktotal - epinfo.ybp -
			cx_palmodes[vmodeidx].vres - 2;
		epinfo.ytotal = cx_palmodes[vmodeidx].vclktotal;
		epinfo.bpp = depth;

		epinfo.clk_src = CLK_EXTERNAL;
		epinfo.automods = cx_palmodes[vmodeidx].automode;
		epinfo.clk_edge = EDGE_FALLING;
		epinfo.pol_blank = POL_LOW;
		epinfo.pol_xsync = POL_LOW,
			epinfo.pol_ysync = POL_LOW;

		epinfo.refresh = cx_palmodes[vmodeidx].clkfrequency;
		epinfo.pixclock = 1000000000 / ( epinfo.refresh / 1000);

		epinfo.configure = cx25871_config;
		epinfo.on = cx25871_on;
		epinfo.off = cx25871_off;
		break;
	}
}

static int
ep93xxfb_alloc_videomem(void)
{
	unsigned long adr,size,pgsize;
	int order;

	pr_debug("%s - enter\n", __FUNCTION__);

	epinfo.fb_log = NULL;
	epinfo.fb_size = PAGE_ALIGN(MAX_FBMEM_SIZE);
	order = get_order(epinfo.fb_size);
	epinfo.fb_log = (void*) __get_free_pages(GFP_KERNEL, order);

	if (epinfo.fb_log) {
		epinfo.fb_phys = __virt_to_phys((int) epinfo.fb_log);
		adr = (unsigned long)epinfo.fb_log;
		size = epinfo.fb_size;
		pgsize = 1 << order;
		do {
			adr += pgsize;
			SetPageReserved(virt_to_page(adr));
		} while(size -= pgsize);
	} else{
		pr_err("%s memory fail\n",__FUNCTION__);
		return -ENOMEM;
	}

	memset(epinfo.fb_log,0x00, epinfo.fb_size);

	pr_debug("   fb_log_addres = 0x%p\n", epinfo.fb_log);
	pr_debug("   fb_phys_address = 0x%x\n", epinfo.fb_phys);
	pr_debug("   fb_size = %lu\n", epinfo.fb_size);
	pr_debug("   fb_page_order = %d\n", order);
	pr_debug("%s - exit\n", __FUNCTION__);

	return 0;
}

static void
ep93xxfb_release_videomem(void)
{
	unsigned long adr,size,psize;
	int order;

	pr_debug("%s - enter\n", __FUNCTION__);

	if (epinfo.fb_log) {
		order = get_order(epinfo.fb_size);
		adr = (unsigned long)epinfo.fb_log;
		size = epinfo.fb_size;
		psize = 1 << order ;
		do {
			adr += psize;
			ClearPageReserved(virt_to_page(adr));
		} while(size -= psize);
		free_pages((unsigned long)epinfo.fb_log, order );
	}

	pr_debug("%s - exit\n", __FUNCTION__);
}

static void
ep93xxfb_setinfo(struct fb_info *info)
{

	info->pseudo_palette = pseudo_palette;
	info->var.xres = epinfo.xres;
	info->var.yres = epinfo.yres;
	info->var.xres_virtual = epinfo.xres;
	info->var.yres_virtual = epinfo.yres;
	info->var.bits_per_pixel = depth;

	ep93xxfb_setcol(info);

	info->var.activate = FB_ACTIVATE_NOW;
	info->var.left_margin = epinfo.xbp;
	info->var.right_margin = epinfo.xfp;
	info->var.upper_margin = epinfo.ybp;
	info->var.lower_margin = epinfo.yfp;
	info->var.hsync_len = epinfo.xsync;
	info->var.vsync_len = epinfo.ysync;

	if( epinfo.pol_xsync == POL_HIGH )
		info->var.sync |= FB_SYNC_HOR_HIGH_ACT;
	if( epinfo.pol_ysync == POL_HIGH )
		info->var.sync |= FB_SYNC_VERT_HIGH_ACT;

	info->var.vmode = FB_VMODE_NONINTERLACED | FB_VMODE_YWRAP;
	info->fix.smem_start = epinfo.fb_phys;
	info->fix.smem_len = epinfo.fb_size;
	info->fix.type = FB_TYPE_PACKED_PIXELS;
	info->fix.line_length = epinfo.xres * (epinfo.bpp / 8);
	info->screen_base = epinfo.fb_log;
	info->var.pixclock = epinfo.pixclock;
	info->fix.ypanstep = 1;
	info->fix.ywrapstep = 1;
}

static int
ep93xxfb_config(struct fb_info *info)
{
	unsigned long attribs;

	pr_debug("%s - enter\n", __FUNCTION__);

	ep93xxfb_get_par(info);

	if (ep93xxfb_alloc_videomem() != 0) {
		pr_err("Unable to allocate video memory\n");
		return -ENOMEM;
	}

	if (ep93xxfb_setclk() != 0) {
		pr_err("Unable to set pixel clock\n");
		ep93xxfb_release_videomem();
		return -ENODEV;
	}

	ep93xxfb_syscon_outl(inl(EP93XX_SYSCON_DEVICE_CONFIG) |
			     EP93XX_SYSCON_DEVICE_CONFIG_RASONP3,
			     EP93XX_SYSCON_DEVICE_CONFIG);

	ep93xxfb_timing_signal_generation();

	/* set video memory parameters */
	outl(epinfo.fb_phys, EP93XX_RASTER_VIDSCRNPAGE);
	outl(epinfo.yres , EP93XX_RASTER_SCRNLINES);
	outl(((epinfo.xres * epinfo.bpp) / 32) - 1, EP93XX_RASTER_LINELENGTH);
	outl((epinfo.xres * epinfo.bpp) / 32, EP93XX_RASTER_VLINESTEP);

	/* set pixel mode */
	ep93xxfb_set_pixelmode(depth);

	attribs = 0;
	attribs |= CONFIG_EP93XX_VIDEOATTRIBS_SDSEL;

	if(epinfo.clk_edge == EDGE_RISING)
		attribs |= EP93XX_RASTER_VIDEOATTRIBS_INVCLK;
	if(epinfo.pol_blank == POL_HIGH)
		attribs |= EP93XX_RASTER_VIDEOATTRIBS_BLKPOL;
	if(epinfo.pol_xsync == POL_HIGH)
		attribs |= EP93XX_RASTER_VIDEOATTRIBS_HSPOL;
	if(epinfo.pol_ysync == POL_HIGH)
		attribs |= EP93XX_RASTER_VIDEOATTRIBS_VCPOL;

	ep93xxfb_raster_outl(attribs, EP93XX_RASTER_VIDEOATTRIBS);
	ep93xxfb_setinfo(info);

	if(epinfo.configure)
		(epinfo.configure)(epinfo.automods);

	ep93xxfb_blank(0, info);

	pr_debug("%s - exit\n", __FUNCTION__);

	return 0;
}

static int
ep93xxfb_ioctl(struct fb_info *info, unsigned int cmd, unsigned long arg)
{
	struct ep93xx_cursor cursor;
	struct ep93xx_blit blit;
	struct ep93xx_fill fill;
	struct ep93xx_line line;
	struct fb_copyarea cparea;
	unsigned long caps;

	switch (cmd) {
	case FBIO_EP93XX_GET_CAPS:
		caps = EP93XX_CAP_CURSOR;
		if (ep93xx_have_graphics_accelerator())
			caps |= EP93XX_CAP_LINE | EP93XX_CAP_FILL |
				EP93XX_CAP_BLIT | EP93XX_CAP_COPY;
		if (copy_to_user((void *)arg, &caps, sizeof(unsigned long)))
			return -EFAULT;
		break;
        case FBIO_EP93XX_CURSOR:
		if (copy_from_user(&cursor, (void *)arg, sizeof(struct ep93xx_cursor)))
			return -EFAULT;
		return ep93xxfb_cursor(info, &cursor);
	case FBIO_EP93XX_LINE:
		if (!ep93xx_have_graphics_accelerator())
			return -EINVAL;
		if (copy_from_user(&line, (void *)arg, sizeof(struct ep93xx_line)))
			return -EFAULT;
		return ep93xxfb_line(info, &line);
	case FBIO_EP93XX_FILL:
		if (!ep93xx_have_graphics_accelerator())
			return -EINVAL;
		if (copy_from_user(&fill, (void *)arg, sizeof(struct ep93xx_fill)))
			return -EFAULT;
		return ep93xxfb_fillrect(info, &fill);
	case FBIO_EP93XX_BLIT:
		if (!ep93xx_have_graphics_accelerator())
			return -EINVAL;
		if (copy_from_user(&blit, (void *)arg, sizeof(struct ep93xx_blit)))
			return -EFAULT;
		return ep93xxfb_imageblit(info, &blit);
        case FBIO_EP93XX_COPY:
		if (!ep93xx_have_graphics_accelerator())
			return -EINVAL;
		if (copy_from_user(&cparea, (void *)arg, sizeof(struct fb_copyarea)))
			return -EFAULT;
		ep93xxfb_copyarea(info, &cparea);
		break;
        default:
		return -EINVAL;
	}

	return 0;
}

static struct fb_ops ep93xxfb_ops = {
	.owner          = THIS_MODULE,
	.fb_setcolreg   = ep93xxfb_setcolreg,
	.fb_check_var   = ep93xxfb_check_var,
	.fb_set_par     = ep93xxfb_set_par,
	.fb_blank       = ep93xxfb_blank,
	.fb_pan_display = ep93xx_pan_display,
	.fb_fillrect	= cfb_fillrect,
	.fb_copyarea	= cfb_copyarea,
	.fb_imageblit   = cfb_imageblit,
	.fb_ioctl       = ep93xxfb_ioctl,
	.fb_mmap        = ep93xxfb_mmap,
};

/*
 * Parse user specified options (`video=ep93xxfb:')
 * example:
 * 	video=ep93xxfb:640x480@60,16bpp,CRT
 */
static int
ep93xxfb_setup(struct platform_device *pdev)
{
	int i, j;
#if !defined(MODULE)
	char *args, *arg;
#endif

#if !defined(MODULE)
	if (fb_get_options("ep93xxfb", &args))
		return -ENODEV;

	pr_debug("%s - %s\n", __FUNCTION__, args);

	while ((arg = strsep(&args, ",")) != NULL) {
		if (strcmp(arg, "8bpp") == 0) {
			depth = 8;
			continue;
		} else if (strcmp(arg, "16bpp") == 0) {
			depth = 16;
			continue;
		} else if (strcmp(arg, "24bpp") == 0) {
			depth = 24;
			continue;
		} else if (strcmp(arg, "32bpp") == 0) {
			depth = 32;
			continue;
		} else if (strcmp(arg, "CRT") == 0) {
			vout = CRT_OUT;
			continue;
		} else if (strcmp(arg, "LCD") == 0) {
			vout = LCD_OUT;
			continue;
		} else if (strcmp(arg, "NTSC") == 0) {
			vout = NTSC_OUT;
			continue;
		} else if (strcmp(arg, "PAL") == 0) {
			vout = PAL_OUT;
			continue;
		} else
			vmode = arg;
	}
#endif

	if (depth == -1)
		depth = CONFIG_FB_EP93XX_DEFAULT_BPP;
	switch (depth) {
	case 8:
		pixel_mask = 3;
		pixel_shift = 2;
		pixel_format = 0x00040000;
		break;
	case 16:
		pixel_mask = 1;
		pixel_shift = 1;
		pixel_format = 0x00080000;
		break;
	case 24:
		pixel_mask = 0;
		pixel_shift = 0;
		pixel_format = 0x00100000;
		break;
	case 32:
		pixel_mask = 0;
		pixel_shift = 0;
		pixel_format = 0x00100000;
		break;
	default:
		dev_err(&pdev->dev, "not support depth: %d\n", depth);
		return -EINVAL;
	}

	if (!vmode)
		vmode = CONFIG_FB_EP93XX_DEFAULT_VIDEOMODE;
	if (vout == -1)
		vout = CONFIG_FB_EP93XX_DEFAULT_OUT;

	switch(vout) {
	case CRT_OUT:
		for (i = 0; i < ep93xxfb_crtmodes_sz; i++) {
			for (j = 0; ep93xxfb_crtmodes[i].name[j]; j++)
				if (!strcmp(vmode, ep93xxfb_crtmodes[i].name[j])) {
					vmodeidx = i;
					break;
				}
		}
		break;
        case LCD_OUT:
		for (i = 0; i < ep93xxfb_lcdmodes_sz; i++) {
			for (j = 0; ep93xxfb_lcdmodes[i].name[j]; j++)
				if (!strcmp(vmode, ep93xxfb_lcdmodes[i].name[j])) {
					vmodeidx = i;
					break;
				}
		}
		depth = 16;
		break;
	case NTSC_OUT:
		for (i = 0; i < cx_ntscmodes_sz; i++) {
			if (!strcmp(vmode, cx_ntscmodes[i].name)) {
				vmodeidx = i;
				break;
			}
		}
		break;
	case PAL_OUT:
		for (i = 0; i < cx_palmodes_sz; i++) {
			if (!strcmp(vmode, cx_palmodes[i].name)) {
				vmodeidx = i;
				break;
			}
		}
		break;
	default:
		dev_err(&pdev->dev, "not support video mode: %d\n", vout);
		return -EINVAL;
	}

	if (vmodeidx == -1) {
		dev_err(&pdev->dev, "not support videomode: %s\n", vmode);
		return -EINVAL;
	}

	return 0;
}

static int __init
ep93xxfb_probe(struct platform_device *pdev)
{
	struct fb_info *info;
	struct resource *mem;
	struct resource *irq;
	int ret;
	int arb;

	pr_debug("%s - enter\n", __FUNCTION__);

	ret = ep93xxfb_setup(pdev);
	if (ret)
		return ret;

	irq = platform_get_resource(pdev, IORESOURCE_IRQ, 0);
	if (!irq) {
		pr_err("Can't find irq resource.\n");
		return -ENODEV;
	}

	mem = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (!mem) {
		pr_err("Can't find memory resource.\n");
		return -ENODEV;
	}

	cursor_data = kmalloc(64*64*2, GFP_KERNEL);
	if (!cursor_data) {
		pr_err("Unable to allocate memory for hw_cursor\n");
		return -ENOMEM;
	}
	memset(cursor_data, 0x00, 64*64*2);

	if (!request_mem_region(mem->start,mem->end - mem->start + 1, FBDEV_NAME))
		return -EBUSY;

	info = framebuffer_alloc(sizeof(struct fb_info), &pdev->dev);
	if(!info) {
		pr_err("Unable to allocate memory for frame buffer\n");
		return -ENOMEM;
	}

	info->flags = FBINFO_DEFAULT;
	strncpy(info->fix.id, FBDEV_NAME, sizeof(info->fix.id));
	info->fix.mmio_start = mem->start;
	info->fix.mmio_len = mem->end - mem->start + 1;
	info->fbops = &ep93xxfb_ops;
	info->pseudo_palette = info->par;
	info->state = FBINFO_STATE_RUNNING;

	ret = fb_alloc_cmap(&info->cmap, 256, 0);
	if (ret < 0)
		goto fbuff;

	ret = ep93xxfb_request_irq(irq->start);
	if (ret != 0) {
		pr_err("Can't get irq %i, err %d\n", irq->start, ret);
                return -EBUSY;
	}

	ret = ep93xxfb_config(info);
	if (ret < 0)
		goto clmap;

	ret = register_framebuffer(info);
	if (ret < 0) {
		pr_err("Unable to register ep93xxfb frame buffer\n");
		goto clmap;
	}

	platform_set_drvdata(pdev, info);

	pr_info("fb%d: EP93xx frame buffer at %dx%dx%dbpp\n", info->node,
		info->var.xres, info->var.yres, info->var.bits_per_pixel);

	register_reboot_notifier(&cx25871_notifier);

	/*change the raster arb to the highest one--Bo*/
	arb = inl(EP93XX_SYSCON_BMAR) & 0x3f8;
	arb |= 0x01;
	outl(arb, EP93XX_SYSCON_BMAR);

	pr_debug("%s - exit\n", __FUNCTION__);

	return 0;

 clmap:
	fb_dealloc_cmap(&info->cmap);

 fbuff:
	framebuffer_release(info);
	return ret;
}

static int
ep93xxfb_remove(struct platform_device *pdev)
{
	struct resource *mem;
	struct resource *irq;
	struct fb_info *info;
	struct ep93xx_cursor cursor;

	pr_debug("%s - enter\n", __FUNCTION__);

	info = platform_get_drvdata(pdev);

	ep93xxfb_release_videomem();

	mem = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	release_mem_region(mem->start, mem->end - mem->start + 1);

	irq = platform_get_resource(pdev, IORESOURCE_IRQ, 0);
	ep93xxfb_free_irq(irq->start);

	platform_set_drvdata(pdev, NULL);
	unregister_framebuffer(info);

	fb_dealloc_cmap(&info->cmap);
	framebuffer_release(info);

	cursor.flags = CURSOR_OFF;
	ep93xxfb_cursor(info, &cursor);

	if (cursor_data != NULL)
		kfree(cursor_data);

	unregister_reboot_notifier(&cx25871_notifier);
	ep93xxfb_blank(1, info);

	pr_debug("%s - exit\n", __FUNCTION__);

	return 0;
}

static struct platform_driver ep93xxfb_driver = {
	.probe  = ep93xxfb_probe,
	.remove = ep93xxfb_remove,
	.driver = {
		.name   = FBDEV_NAME,
	},
};

static int __init
ep93xxfb_init(void)
{
	pr_debug("%s\n", __FUNCTION__);

	return platform_driver_register(&ep93xxfb_driver);
}

static void __exit
ep93xxfb_exit(void)
{
	pr_debug("%s\n", __FUNCTION__);

	platform_driver_unregister(&ep93xxfb_driver);
}

module_init(ep93xxfb_init);
module_exit(ep93xxfb_exit);

module_param(vmode, charp, 0);
MODULE_PARM_DESC(vmode, "Video mode");

module_param(vout, int, 0);
MODULE_PARM_DESC(vout,"Specify video output (0 = CRT, 1 = LCD, 2 = NTSC, 3 = PAL)");

module_param(depth, int, 16);
MODULE_PARM_DESC(depth ,"Color depth (8,16,24,32)");

MODULE_LICENSE("GPL");

